@@ -3,10 +3,6 @@
 """
 `Span` class
 
-`Span` sub-classes the `string` class in Python, thus enabling basic usages of
-string (as e.g. isupper, lower, ... see 
-https://docs.python.org/3.8/library/string.html)
-
 Its main original methods are
  - `partition(range(start,stop))` : which partitions the initial `Span` in three
  new `Span` instance, collected in a unique `Spans` instance (see below)
@@ -16,49 +12,50 @@ Its main original methods are
 `start` to position `stop` by `step` in sub-strings of size `size`, all grouped 
 in a list of Span objects
 
+In addition, one can compare two `Span` using the non-overlapping order `<` and `>` or the overlapping-order `<=` and `>=` when two `Span`s overlap.
+
+Finally, since a `Span` can be seen as a selection of a set of character positions from the parent string, one can apply the basic set operations to two `Span`s, in order to construct more elaborated `Span` instance. 
+
 """
 
+from hashlib import sha1
 from .tools import _checkRange, _checkRanges, _startstop, _removeRange
 from .tools import _checkSpan, _checkSameString
 from .tools import _combineRanges, _cutRanges
 
 class Span():
     """
-Pseudo-Subclass of the Python string class.
-It allows manipulating a string as a usual Python object, excepts it 
-returns Span instances. Especially for Span instances are the 
-methods : 
- - `isupper()` (for instance), which tells whether the string is 
- uppercase or not
- - `lower()` or `upper()`, to make the `Span` lower-/upper-case.
-See more string methods on [the Python standard library documentation](https://docs.python.org/3/library/stdtypes.html#text-sequence-type-str).
-
+`Span` object is basically a collection of a parent string named `string`, and a `ranges` collection (list of range) of positions. Its basic usage is its `str` method, which consists in a string extracted from all intervals defined in the `ranges` list, and joined by the `subtoksep` separator.
     """ 
     
     def __init__(self,
                  string='',
-                 ranges=[],
-                 subtoksep=chr(32),):
+                 ranges=None,
+                 subtoksep=chr(32),
+                 encoding='utf-8'):
         """
-`Span` object is basically a `string` with a `ranges` (list of range) position. Its basic usage is : 
- - `string` extracted from all intervals defined in the `ranges` list
-and its attributes are
- - `Span.string`	 -> a string
- - `Span.ranges`	 -> a list of ranges
- - `Span.subtoksep` -> a string, preferably of length 1
+`Span` attributes are
+ - `Span.string`	 -> a string, default is empty
+ - `Span.ranges`	 -> a list of ranges, default is None, in which case the ranges are calculated to contain the entire string
+ - `Span.subtoksep`  -> a string, preferably of length 1, default is a white space ' '
+ - `Span.encoding`   -> a string, representing the encoding, default is `'utf-8'`
          """
         self.string = str(string)
-        # self._extra_attributes = set(_Span_methods)
         self.subtoksep = str(subtoksep)
+        self.encoding = str(encoding)
+        # construct the ranges only when it is not passed as argument
+        # keep empty list in case it is a list
+        if ranges is None:
+            self.ranges = [range(len(self.string)),] if len(self.string) else []
+            return None
+        self.ranges = ranges
         # check if ranges are correct
-        _checkRanges(ranges)
+        _checkRanges(self.ranges)
         # correct ranges in case some are over the string size
         ranges_ = [range(*_startstop(self.string,r.start,r.stop))
-                   for r in ranges]
+                   for r in self.ranges]
         # withdraw overlapping ranges
         self.ranges = _combineRanges(ranges_)
-        if not self.ranges:
-            self.ranges.append(range(len(self.string)))
         return None
 
     def _append_range(self, r):
@@ -73,14 +70,14 @@ and its attributes are
         """
 Append a range object to self.ranges. The range r must be given in absolute coordinates.
 Return self (append in place).
-Raise a ValueError in case r is not a range object.
-Raise a Warning in case r has start or stop attributes outside the size of Span.string, in which case thse parameters are recalculated to fit Span.string (being either 0 for start or len(Span.string) for stop).
+Raise a `ValueError` in case r is not a range object.
+Raise a `BoundaryWarning` in case r has start or stop attributes outside the size of Span.string, in which case thse parameters are recalculated to fit Span.string (being either 0 for start or len(Span.string) for stop).
         """
         self._append_range(r)
         self.ranges = [r for r in _combineRanges(self.ranges) if r]
         return self
     
-    def append_ranges(self,ranges):
+    def append_ranges(self, ranges):
         """
 Append a list of range objects to self.ranges. This method applies `append_range` several times, so please see its documentation for more details.
         """
@@ -116,10 +113,10 @@ Remove a list of range objects toself.ranges. This method applies `remove_range`
         """Return the length of the string associated with the Span"""
         l = sum(r.stop-r.start for r in self.ranges) 
         l += (len(self.ranges)-1)*len(self.subtoksep)
-        return l if l>0 else 0 # in case there is no ranges
+        return max(l,0) # in case ranges=[], l<0
     
     def __repr__(self):
-        """Return the two main arguments (namely the `string` and the number of ranges) of a `Span` instance in a readable way."""
+        """Return the two main elements (namely the `string` and the `ranges` attributes) of a `Span` instance in a readable way."""
         mess = "Span('{}', ".format(str(self))
         mess += '['+','.join('('+str(r.start)+','+str(r.stop)+')' 
                              for r in self.ranges)
@@ -128,10 +125,19 @@ Remove a list of range objects toself.ranges. This method applies `remove_range`
     
     def __str__(self):
         """
-`str(Span)` method returns the recombination of the extract of each `Span.subSpan` from the `Span.string` attribute corresponding to all its `Span.ranges` attribute.
+`str(Span)` method returns the recombination of the extract of each `Span.subSpan` from the `Span.string` attribute corresponding to all its `Span.ranges` attribute, and joined by the `Span.subtoksep` character.
         """
         return self.subtoksep.join(self.string[r.start:r.stop] 
                                    for r in self.ranges)
+
+    def __hash__(self,):
+        """
+Make the Span object hashable, such that it can serve for set and dict.keys. Span is constructed on the unicity of the Span object, that is, this is the hash of the string made of the parent string, plus the string representation of the instance, including subtoksep. Everything is then converted to hashlib.sha1.hexdigest
+        """
+        _bytes_ = bytes(repr(self) + self.string + self.subtoksep,
+                        encoding=self.encoding)
+        return int(sha1(_bytes_).hexdigest(), 16)
+        
     
     def __contains__(self,s):
         """If the object to be compared with is a Span related to the same string as this instance, check whether the ranges are overlapping. Otherwise, check whether the string str(s) (which transforms the other Span instance in a string in case s is not related to the same string) is a sub-string of the `Span` instance."""
@@ -163,19 +169,19 @@ Note: As for the usual Python string, a slice with positions outside str(Span) w
     def get_subSpan(self,n):
         """
 Get the Span associated to the ranges elements n (being an integer or a slice).
-Return a Span.
-Raise a IndexError in case n is larger than the number of ranges in self.ranges.
+Return a Span object.
+Raise an IndexError in case n is larger than the number of ranges in self.ranges.
         """
-        span = Span(string=self.string,
-                    ranges=[self.ranges[n],],
-                    subtoksep=self.subtoksep)          
+        span = self.__class__(string=self.string,
+                              ranges=[self.ranges[n],],
+                              subtoksep=self.subtoksep)          
         return span
     
     @property
     def subSpans(self,):
         """
 Get the Span associated to each Span.ranges in a Span object.
-Return a Spans object. Keep the attributes in case Span.carry_attributes is True. 
+Return a Span object.  
         """
         return [self.get_subSpan(i) for i in range(len(self.ranges))]
 
@@ -193,19 +199,19 @@ Raise a ValueError when one object is not a Span instance
                  span.subtoksep==self.subtoksep,]
         return all(bools)
 
-    def __add__(self,span):
+    def __add__(self, span):
         """If the two Span objects have same strings, returns a new Span object with combined ranges of the initial ones."""
         return self.union(span)
 
-    def __sub__(self,span):
+    def __sub__(self, span):
         """If the two Span objects have same strings, returns a new Span object with ranges of self with Span ranges removed. Might returns an empty Span."""
         return self.difference(span)
 
-    def __mul__(self,span):
+    def __mul__(self, span):
         """If the two Span objects have same strings, returns a new Span object with ranges of self having intersection with Span ranges removed. Might returns an empty Span."""
         return self.intersection(span)
 
-    def __truediv__(self,span):
+    def __truediv__(self, span):
         """If the two Span objects have same strings, returns a new Span object with ranges of self having symmetric_difference with Span ranges removed. Might returns an empty Span."""
         return self.symmetric_difference(span)
     
@@ -217,6 +223,30 @@ Raise a ValueError when one object is not a Span instance
     def stop(self,):
         """Returns the ending position (an integer) of the last ranges. Make sense only for contiguous Span."""
         return self.ranges[-1].stop
+    
+    def __lt__(self, span):
+        """Returns True if Span is entirely on the left of span (the Span object given as parameter). Make sense only for contiguous Span."""
+        _checkSpan(span)
+        _checkSameString(self, span)
+        return self.stop < span.start 
+    
+    def __gt__(self, span):
+        """Returns True if Span is entirely on the right of span (the Span object given as parameter). Make sense only for contiguous Span."""
+        _checkSpan(span)
+        _checkSameString(self, span)
+        return span.stop < self.start
+    
+    def __le__(self, span):
+        """Returns True if Span is partly on the left of span (the Span object given as parameter). Make sense only for contiguous Span."""
+        _checkSpan(span)
+        _checkSameString(self, span)
+        return self.start <= span.start <= self.stop
+    
+    def __ge__(self, span):
+        """Returns True if Span is partly on the right of span (the Span object given as parameter). Make sense only for contiguous Span."""
+        _checkSpan(span)
+        _checkSameString(self, span)
+        return span.start < self.start <= span.stop
 
     def union(self, span):
         """
@@ -236,10 +266,21 @@ Takes a Span object as entry, and returns a new Span instance, with Span.ranges
 | TypeError | in case the span.string is not the same as Span.string. |
         """
         _checkSpan(span)
-        _checkSameString(self,span,'Span')
-        newSpan = Span(string=self.string,
-                       subtoksep=self.subtoksep,
-                       ranges=self.ranges).append_ranges(span.ranges)
+        _checkSameString(self,span)
+        # in case there is no ranges, the constructor would fill them
+        if not self and span:
+            newSpan = self.__class__(string=self.string,
+                                     subtoksep=self.subtoksep,
+                                     ranges=span.ranges)
+        elif not self and not span:
+            newSpan = self.__class__(string=self.string,
+                                     subtoksep=self.subtoksep,
+                                     ranges=span.ranges)
+            newSpan.ranges = []
+        else:
+            newSpan = self.__class__(string=self.string,
+                                     subtoksep=self.subtoksep,
+                                     ranges=self.ranges).append_ranges(span.ranges)
         return newSpan
 
     def difference(self, span):
@@ -260,13 +301,13 @@ Takes a Span object as entry, and returns a new Span instance with Span.ranges g
 | TypeError | in case the span.string is not the same as Span.string. |
         """
         _checkSpan(span)
-        _checkSameString(self,span,'Span')
-        newSpan = Span(string=self.string,
-                       subtoksep=self.subtoksep,
-                       ranges=self.ranges).remove_ranges(span.ranges)
+        _checkSameString(self,span)
+        newSpan = self.__class__(string=self.string,
+                                 subtoksep=self.subtoksep,
+                                 ranges=self.ranges).remove_ranges(span.ranges)
         return newSpan
     
-    def intersection(self,span):
+    def intersection(self, span):
         """
 Takes a Span object as entry and returns a new Span whose Span.ranges given by the intersection of the actual Span.ranges with the span.ranges, when one sees the `ranges` attributes as sets of positions of each instance..
 
@@ -284,30 +325,45 @@ Takes a Span object as entry and returns a new Span whose Span.ranges given by t
 | TypeError | in case the span.string is not the same as Span.string. |
         """
         _checkSpan(span)
-        _checkSameString(self,span,'Span')
+        _checkSameString(self,span)
+        # if span is empty, the following procedure doesn't work
+        # one has to construct an empty ranges attribute by hand
+        # since the initialization of Span constructs it for us otherwise
+        if not span.ranges:
+            newSpan = self.__class__(string=self.string,
+                                     subtoksep=self.subtoksep,
+                                     ranges=self.ranges)
+            newSpan.ranges = span.ranges
+            return newSpan
         # one uses the fact that AxB = A+B-(A-B)-(B-A)
-        AmB = Span(string=self.string,
-                   subtoksep=self.subtoksep,
-                   ranges=self.ranges).remove_ranges(span.ranges)
-        BmA = Span(string=self.string,
-                   subtoksep=self.subtoksep,
-                   ranges=span.ranges).remove_ranges(self.ranges)
-        newSpan = Span(string=self.string,
-                       subtoksep=self.subtoksep,
-                       ranges=span.ranges).append_ranges(self.ranges)
-        newSpan.remove_ranges(BmA.ranges)
+        # A-B
+        AmB = self.__class__(string=self.string,
+                             subtoksep=self.subtoksep,
+                             ranges=self.ranges).remove_ranges(span.ranges)
+        # B-A
+        BmA = self.__class__(string=self.string,
+                             subtoksep=self.subtoksep,
+                             ranges=span.ranges).remove_ranges(self.ranges)
+        # A+B
+        newSpan = self.__class__(string=self.string,
+                                 subtoksep=self.subtoksep,
+                                 ranges=span.ranges).append_ranges(self.ranges)
+        # A+B - (A-B)
         newSpan.remove_ranges(AmB.ranges)
-        ranges = [r for r in _combineRanges(newSpan.ranges) if r]
-        if not ranges:
-            newSpan = Span()
-            newSpan.ranges = ranges
-        else:
-            newSpan.ranges = ranges
+        # A+B - (A-B) - (B-A)
+        newSpan.remove_ranges(BmA.ranges)
+        ranges_ = [r for r in _combineRanges(newSpan.ranges) if r]
+        newSpan = self.__class__(string=self.string,
+                                 subtoksep=self.subtoksep,
+                                 ranges=ranges_)
+        # withdraw ranges attribute in case it's an empty list
+        if not ranges_:
+            newSpan.ranges = ranges_
         return newSpan
     
-    def symmetric_difference(self,span):
+    def symmetric_difference(self, span):
         """
-Takes a Span object as entry, and return a new Span instance whose Span.ranges given by the symmetric difference of the actual Span.ranges with the span.ranges, when one sees the `ranges` attributes as sets of positions of each instance..
+Takes a Span object as entry, and return a new Span instance whose Span.ranges given by the symmetric difference of the actual Span.ranges with the span.ranges, when one sees the `ranges` attributes as sets of positions of each instance.
 
 | Parameters | Type | Details |
 | --- | --- | --- | 
@@ -323,77 +379,83 @@ Takes a Span object as entry, and return a new Span instance whose Span.ranges g
 | TypeError | in case the span.string is not the same as Span.string. |
         """
         _checkSpan(span)
-        _checkSameString(self,span,'Span')
+        _checkSameString(self,span)
         # one uses the fact that A/B = (A-B)+(B-A)
-        AmB = Span(string=self.string,
-                   subtoksep=self.subtoksep,
-                   ranges=self.ranges).remove_ranges(span.ranges)
-        BmA = Span(string=self.string,
-                   subtoksep=self.subtoksep,
-                   ranges=span.ranges).remove_ranges(self.ranges)
-        newSpan = Span(string=self.string,
-                       subtoksep=self.subtoksep,
-                       ranges=AmB.ranges).append_ranges(BmA.ranges)
-        newSpan.ranges = [r for r in _combineRanges(newSpan.ranges) if r]
+        # A-B
+        AmB = self.__class__(string=self.string,
+                             subtoksep=self.subtoksep,
+                             ranges=self.ranges).remove_ranges(span.ranges)
+        # B-A
+        BmA = self.__class__(string=self.string,
+                             subtoksep=self.subtoksep,
+                             ranges=span.ranges).remove_ranges(self.ranges)
+        # does the union of ranges: (A-B)+(B-A)
+        ranges_ = []
+        ranges_.extend(AmB.ranges)
+        ranges_.extend(BmA.ranges)
+        ranges_ = [r for r in _combineRanges(ranges_) if r]
+        newSpan = self.__class__(string=self.string,
+                                 subtoksep=self.subtoksep,
+                                 ranges=ranges_)
+        # withdraw ranges attribute in case it's an empty list
+        # since the initialization of Span constructs it for us otherwise
+        if not ranges_:
+            newSpan.ranges = ranges_
         return newSpan
       
     def _prepareSpans(self,ranges,remove_empty):
         """Utility that removes empty ranges and constructs a list of Span objects."""
         if remove_empty:
             ranges = [r for r in ranges if any(r_ for r_ in r) or len(r)==2]
-            # or len==2: to handle the case containing only a subtoksep
-        spans = [Span(string=self.string,ranges=r,subtoksep=self.subtoksep)
+        spans = [self.__class__(string=self.string,
+                                ranges=r,
+                                subtoksep=self.subtoksep)
                   for r in ranges]
         return spans
     
-    def partition(self,start,end,remove_empty=False):
+    def partition(self, start, stop, remove_empty=False):
         """
 Split the `Span.string` in three `Span` objects : 
  - `string[:start]`
  - `string[start:stop]`
  - `string[stop:]`
-and put all non-empty `Span` objects in a `Spans` instance.
+and put all non-empty `Span` objects in a list of `Span` instances.
 
 It acts a bit like the `str.partition(s)` method of the Python
-`string` object, but `partition_Span` takes `start` and `end` 
-argument instead of a string. So in case one wants to split a string in three 
-sub-strings using a string 's', use `Span.partition(s)` instead, 
-inherited from `str.partition(s)`.
-
-NB : `Span.partition(s)` has no `non_empty` option.
+`string` object, but `Span.partition` takes `start` and `stop` 
+argument instead of a string. 
 
 | Parameters | Type | Details |
 | --- | --- | --- | 
 | `start` | int | Starting position of the splitting sequence. |
-| `end` | int | Ending position of the splitting sequence. |
+| `stop` | int | Ending position of the splitting sequence. |
 | `remove_empty` | bool. Default is `False` | If `True`, returns a `list of Span` instance with only non-empty `Span` objects. see `__bool__()` method for non-empty `Span` |
 
 | Returns | Type | Details |
 | --- | --- | --- | 
 | `spans` | `list` of `Span` objects | The `list` object containing the different `Span` objects. |
         """
-        start,end = _startstop(self,start,end)
-        # ranges = _splitRanges(self.ranges,start,end,step=len(self.subtoksep))
+        start,stop = _startstop(self,start,stop)
         r1,temp_ranges = _cutRanges(self.ranges,start,step=len(self.subtoksep))
-        r2,r3 = _cutRanges(temp_ranges,end-start,step=len(self.subtoksep))
+        r2,r3 = _cutRanges(temp_ranges,stop-start,step=len(self.subtoksep))
         ranges = [r1,r2,r3]
         spans = self._prepareSpans(ranges,remove_empty)
         return spans
     
-    def split(self,cuts,remove_empty=False):
+    def split(self, cuts, remove_empty=False):
         """
 Split a text as many times as there are range entities in the cuts list.
-Return a `Spans` instance.
+Return a list of `Span` instances.
 
 This is a bit like `str.split(s)` method from Python `string`
 object, except one has to feed `Span.split` with a full list
-of `range(start,end)` range objects instead of the string 's' in `str.split(s)`
-If the `range(start,end)` tuples in cuts are given by a regex re.finditer
+of `range(start,stop)` range objects instead of the string 's' in `str.split(s)`
+If the `range(start,stop)` tuples in cuts are given by a regex re.finditer
 search on `str(Span)`, the two methods give the same thing. 
 
 | Parameters | Type | Details |
 | --- | --- | --- | 
-| `cuts` | a list of `range(start,end,)` range objects. start/end are integer | Basic usage is to take these cuts from [`re.finditer`](https://docs.python.org/3/library/re.html#re.finditer). The start/end integers are given in the relative coordinate system, that is, in terms of the position in `str(Span)`. |
+| `cuts` | a list of `range(start,stop,)` range objects. start/stop are integer | Basic usage is to take these cuts from [`re.finditer`](https://docs.python.org/3/library/re.html#re.finditer). The start/end integers are given in the relative coordinate system, that is, in terms of the position in `str(Span)`. |
 | `remove_empty` | bool. Default is `False` | If `True`, returns a `list of Span` instance with only non-empty `Span` objects. see `__bool__()` method for non-empty `Span` |
 
 | Return | Type | Details |
@@ -418,18 +480,18 @@ search on `str(Span)`, the two methods give the same thing.
         spans = self._prepareSpans(split_ranges, remove_empty)      
         return spans
 
-    def slice(self,start=0,stop=None,size=1,step=1,remove_empty=False):
+    def slice(self, start=0, stop=None, size=1, step=1, remove_empty=False):
         """
 Cut the `Span.string` in overlapping sequences of strings of size `size` by `step`,
 put all these sequences in separated `Span` objects, and finally 
-put all theses objects in a `Spans` instance.
+put all theses objects in a list of `Span` instances.
 
 | Parameters | Type | Details |
 | --- | --- | --- | 
 | `start` | int | The relative position where to start slicing the Span. |
 | `stop` | int | The relative position where to stop slicing the Span. |
 | `size` | int | The size of the string in each subsequent Span objects. |
-| `step` | int | The number of characters skipped from one Span object to the next one. A character is given by `str(Span)` (relative coordinate) |
+| `step` | int | The number of characters skipped from one Span object to the next one. A character is given by `str(Span)` (called relative position) |
 
 | Returns | Type | Details |
 | --- | --- | --- | 
